@@ -1,14 +1,21 @@
 # Copyright 2021 VMware, Inc.
 # SPDX-License-Identifier: BSD-2
 import abc
+import cgi
+import collections
 import configparser
 import datetime
+import hashlib
 import io
 import logging
+import weakref
 from typing import Any
+from typing import Callable
 from typing import Dict
+from typing import Iterable
 from typing import List
 from typing import Optional
+from typing import TypeVar
 from typing import Union
 
 import more_itertools
@@ -17,24 +24,31 @@ import tau_clients
 from tau_clients import exceptions
 
 
+AbstractClientSubType = TypeVar("AbstractClientSubType", bound="AbstractClient")
+
+
+MultiClientSubType = TypeVar("MultiClientSubType", bound="MultiClientMixin")
+
+
+Disposition = collections.namedtuple("Disposition", ["type", "params"])
+
+
 class AbstractClient(abc.ABC):
-    """"A very basic HTTP client providing basic functionality."""
+    """A very basic HTTP client providing basic functionality."""
 
-    __metaclass__ = abc.ABCMeta
-
-    # Empty tuple, to be overriden by the subclass
+    # Empty tuple, to be overridden by the subclass
     MODULES = ()
 
-    # Empty tuple, to be overriden by the subclass
+    # Empty tuple, to be overridden by the subclass
     FORMATS = ()
 
-    # Default FMT used by the API
-    DATETIME_FMT = "%Y-%m-%d %H:%M:%S"
+    # Empty dictionary, to be overridden by the subclass
+    HOSTED_URLS = {}
 
     @classmethod
     def _get_login_params(
         cls, conf: configparser.ConfigParser, section_name: str
-    ) -> dict:
+    ) -> Dict[str, str]:
         """
         Get the login parameters from a 'ConfigParser' object.
 
@@ -59,16 +73,42 @@ class AbstractClient(abc.ABC):
         else:
             raise configparser.NoOptionError("username", section_name)
 
+    @classmethod
+    def from_conf_all(
+        cls, conf: configparser.ConfigParser, section_name: str
+    ) -> Dict[str, AbstractClientSubType]:
+        """
+        Get ALL the clients covering the whole hosted infrastructure.
+
+        :param configparser.ConfigParser conf: the conf object
+        :param str section_name: the section name
+        :rtype: dict[str, AbstractClientSubtype]
+        :return: the initialized clients indexed by data center label
+        :raises configparser.Error: any exception that can be raised by ConfigParser
+        """
+        login_params = cls._get_login_params(conf, section_name)
+        timeout = conf.getint(section_name, "timeout", fallback=60)
+        verify_ssl = conf.getboolean(section_name, "verify_ssl", fallback=True)
+        return {
+            data_center: cls(
+                api_url=cls.HOSTED_URLS[data_center],
+                login_params=login_params,
+                timeout=timeout,
+                verify_ssl=verify_ssl,
+            )
+            for data_center, data_center_url in cls.HOSTED_URLS.items()
+        }
+
     @classmethod
     def from_conf(
         cls, conf: configparser.ConfigParser, section_name: str
-    ) -> "AbstractClient":
+    ) -> AbstractClientSubType:
         """
         Get the client from a 'ConfigParser' object.
 
         :param configparser.ConfigParser conf: the conf object
         :param str section_name: the section name
-        :rtype: Any
+        :rtype: AbstractClientSubtype
         :return: the initialized client
         :raises configparser.Error: any exception that can be raised by ConfigParser
         """
@@ -96,7 +136,7 @@ class AbstractClient(abc.ABC):
         :param str api_url: the URL of the API endpoint
         :param dict[str, str]: the login parameters
         :param int timeout: the timeout
-        :param boolean verify_ssl: whether to verify the SSL certificate
+        :param bool verify_ssl: whether to verify the SSL certificate
         """
         self._url = api_url
         self._login_params = login_params
@@ -118,9 +158,23 @@ class AbstractClient(abc.ABC):
         """Return whether we have an active session."""
         return self._session is not None
 
-    def _build_url(
-        self, module: str, function: Union[List[str], str], fmt: str = "json"
-    ) -> str:
+    @staticmethod
+    def _parse_disposition(response: requests.Response) -> Optional[Disposition]:
+        """
+        Return the content disposition of a response if present.
+
+        :param Response response: the response
+        :rtype: Disposition|None
+        :return: an object with the content disposition or None.
+        """
+        content_disposition = response.headers.get("content-disposition")
+        if content_disposition:
+            disp_type, disp_params = cgi.parse_header(content_disposition)
+            if disp_type:
+                return Disposition(type=disp_type.lower(), params=disp_params)
+        return None
+
+    def _build_url(self, module: str, function: Union[List[str], str], fmt: str = "json") -> str:
         """
         Build a URL composed by module, function(s), and format.
 
@@ -133,7 +187,7 @@ class AbstractClient(abc.ABC):
         """
         if module not in self.MODULES:
             raise exceptions.InvalidArgument(module)
-        if fmt not in self.FORMATS:
+        if fmt and fmt not in self.FORMATS:
             raise exceptions.InvalidArgument(fmt)
         if isinstance(function, list):
             parts = function
@@ -141,19 +195,27 @@ class AbstractClient(abc.ABC):
             parts = [function] if function else []
         if parts and not any(isinstance(x, str) for x in parts):
             raise exceptions.InvalidArgument(function)
-        return "/".join([self.base, module] + parts) + ".{}".format(fmt)
+        if fmt:
+            return "/".join([self.base, module] + parts) + ".{}".format(fmt)
+        else:
+            return "/".join([self.base, module] + parts)
 
     @classmethod
-    def _parse_response(cls, response: requests.Response) -> Union[dict, str]:
+    def _parse_response(
+        cls, response: requests.Response, raw: bool = False
+    ) -> Union[dict, bytes]:
         """
         Parse the response.
 
         :param requests.Response response: the response
+        :param bool raw: whether the raw response should be parsed
         :rtype: dict|str
         :return: the decoded response
         :raises ApiError: if the response has an error
         """
         try:
+            if raw:
+                return response.content
             ret = response.json()
             if "success" not in ret:
                 raise exceptions.ApiError("No success field in response")
@@ -167,18 +229,22 @@ class AbstractClient(abc.ABC):
 
     @classmethod
     def _parse_exception(
-        cls, request_exception: requests.RequestException, response: requests.Response
+        cls,
+        request_exception: requests.RequestException,
+        response: requests.Response,
+        raw: bool = False,
     ) -> exceptions.ApiError:
         """
         Convert an exception into a ApiError exception.
 
         :param requests.RequestException request_exception: the exception that has been raised
         :param requests.Response response: the response object
+        :param bool raw: whether the raw body should be parsed
         :rtype: ApiError
         :return: the converted exception with as much debug info as possible
         """
         try:
-            cls._parse_response(response)
+            cls._parse_response(response, raw)
         except exceptions.ApiError as api_exception:
             error_message = "{} ({})".format(api_exception, request_exception)
         else:
@@ -188,23 +254,33 @@ class AbstractClient(abc.ABC):
     @classmethod
     def _handle_response(
         cls, response: requests.Response, raw: bool = False
-    ) -> Union[dict, str]:
+    ) -> Union[dict, bytes, tau_clients.NamedBytesIO]:
         """
         Check a response for issues and parse the return.
 
         :param requests.Response response: the response
         :param bool raw: whether the raw body should be returned
-        :rtype: str|dict
+        :rtype: bytes|dict|NamedBytesIO
         :return: the response content
         :raises ApiError: in case of any error
         """
         try:
             response.raise_for_status()
         except requests.RequestException as exception:
-            raise cls._parse_exception(exception, response) from exception
+            raise cls._parse_exception(exception, response, raw) from exception
 
         if raw:
-            return response.content
+            disposition = cls._parse_disposition(response)
+            if disposition and disposition.type == "attachment":
+                sha256_hash = hashlib.sha256()
+                sha256_hash.update(response.content)
+                return tau_clients.NamedBytesIO(
+                    content=response.content,
+                    name=sha256_hash.hexdigest(),  # disposition.params.get("filename")
+                    mime_type=response.headers.get("content-type"),
+                )
+            else:
+                return response.content
         else:
             return cls._parse_response(response)
 
@@ -212,31 +288,29 @@ class AbstractClient(abc.ABC):
         self,
         method: str,
         module: str,
-        function: Union[List[str], str],
+        function: Union[List[str], str, None],
         params: Optional[Dict[str, str]] = None,
         data: Optional[Dict[str, str]] = None,
         headers: Optional[Dict[str, str]] = None,
         files: Optional[Dict[str, str]] = None,
-        raw: bool = False,
-        raw_response: bool = False,
         stream_response: bool = False,
-        fmt: str = "json",
-    ) -> Union[dict, str]:
+        fmt: Optional[str] = "json",
+        raw: bool = False,
+    ) -> Union[dict, bytes, tau_clients.NamedBytesIO]:
         """
         Do the request (and authenticate first if needed).
 
         :param str method: the HTTP request method
         :param str module: the module to contact
-        :param list[str]|str function: a list of string (or a string) defining the route
+        :param list[str]|str|None function: a list of string (or a string) defining the route
         :param dict[str, str] params: a dictionary of parameters
         :param dict[str, str] data: used for POSTs
         :param dict[str, str] headers: headers
-        :param dict[str, str] files: when downloading a file
-        :param str fmt: the requested format
-        :param bool raw: whether to return a non-JSON decoded response
-        :param bool raw_response: whether to return a raw Response object
+        :param dict[str, any] files: when downloading a file
         :param bool stream_response: whether to download immediately
-        :rtype: str|dict
+        :param str|None fmt: the requested format
+        :param bool raw: whether to return a non-JSON decoded response
+        :rtype: bytes|dict|NamedBytesIO
         :return: the response content
         :raises CommunicationError: if there was an error connecting to the resource
         :raises ApiError: if there was on error on the server side
@@ -258,21 +332,19 @@ class AbstractClient(abc.ABC):
             )
         except requests.RequestException as e:
             raise exceptions.CommunicationError(str(e)) from e
-
-        if raw_response:
-            return response
         else:
             return self._handle_response(response, raw)
 
     def post(
         self,
         module: str,
-        function: Union[List[str], str],
+        function: Union[List[str], str, None],
         params: Optional[Dict[str, str]] = None,
         data: Optional[Dict[str, str]] = None,
-        files: Optional[Dict[str, str]] = None,
-        fmt: str = "json",
-    ) -> Union[dict, str]:
+        files: Optional[Dict[str, Any]] = None,
+        fmt: Optional[str] = "json",
+        raw: bool = False,
+    ) -> Union[dict, bytes]:
         """Utility method to issue a POST request (see '_request' for documentation)."""
         return self._request(
             method="POST",
@@ -282,6 +354,7 @@ class AbstractClient(abc.ABC):
             data=data,
             files=files,
             fmt=fmt,
+            raw=raw,
         )
 
     def get(
@@ -289,11 +362,17 @@ class AbstractClient(abc.ABC):
         module: str,
         function: Union[List[str], str],
         params: Optional[Dict[str, str]] = None,
-        fmt: str = "json",
-    ) -> Union[dict, str]:
+        fmt: Optional[str] = "json",
+        raw: bool = False,
+    ) -> Union[dict, bytes, tau_clients.NamedBytesIO]:
         """Utility method to issue a GET request (see '_request' for documentation)."""
         return self._request(
-            method="GET", module=module, function=function, params=params, fmt=fmt
+            method="GET",
+            module=module,
+            function=function,
+            params=params,
+            fmt=fmt,
+            raw=raw,
         )
 
 
@@ -304,15 +383,15 @@ class PortalClient(AbstractClient):
 
     FORMATS = ("json",)
 
+    HOSTED_URLS = tau_clients.NSX_DEFENDER_PORTAL_URLS
+
     def _login(self) -> None:
         """Implement interface (portal client relies on 'username' and 'password')."""
         if self._session is None:
             self._session = requests.sessions.session()
         self.post("login", function=None, data=self._login_params)
 
-    def get_tasks_from_knowledgebase(
-        self, query_string: str, **kwargs
-    ) -> List[Dict[str, Any]]:
+    def get_tasks_from_knowledgebase(self, query_string: str, **kwargs) -> List[Dict[str, Any]]:
         """
         Query tasks from the knowledgebase, most recent first.
 
@@ -348,9 +427,7 @@ class PortalClient(AbstractClient):
         )
         sorted_malscape_tasks = sorted(
             unique_malscape_tasks,
-            key=lambda x: datetime.datetime.strptime(
-                x["submission"], self.DATETIME_FMT
-            ),
+            key=lambda x: tau_clients.parse_datetime(x["submission"]),
             reverse=True,
         )
         return sorted_malscape_tasks
@@ -372,11 +449,11 @@ class PortalClient(AbstractClient):
 
     def get_result(self, uuid: str) -> Dict[str, Any]:
         """
-        Get report results for a given task.
+        Get the result for a given task.
 
         :param str uuid: the unique identifier of the submitted task
         :rtype: dict[str, any]
-        :return: a dictionary containing the analysis report
+        :return: the result of an analysis
         """
         params = tau_clients.purge_none({"uuid": uuid, "report_format": "json"})
         return self.get("analysis", "get_result", params=params)
@@ -394,7 +471,7 @@ class PortalClient(AbstractClient):
         :param str url: the url to analyze
         :param str|None referer: the referer
         :param str|None user_agent: the user agent
-        :param boolean bypass_cache: bypass_cache
+        :param bool bypass_cache: bypass_cache
         :rtype: dict[str, any]
         :return: a dictionary like:
             {
@@ -443,8 +520,8 @@ class PortalClient(AbstractClient):
         :param str|None file_name: if set, represents the name of the file to submit
         :param str|None password: if set, use it to extract the sample
         :param str|None analysis_env: if set, e.g windowsxp
-        :param boolean allow_network_traffic: if set to False, deny network connections
-        :param boolean bypass_cache: whether to re-process a file (requires special permissions)
+        :param bool allow_network_traffic: if set to False, deny network connections
+        :param bool bypass_cache: whether to re-process a file (requires special permissions)
         :rtype: dict[str, any]
         :return: a dictionary like:
             {
@@ -481,19 +558,19 @@ class PortalClient(AbstractClient):
 
 
 class AnalysisClient(AbstractClient):
-    """"Simple client to query the analysis API (Malscape)."""
+    """Simple client to query the analysis API (Malscape)."""
 
     MODULES = ("analysis", "authentication")
 
     FORMATS = ("json", "xml")
 
+    HOSTED_URLS = tau_clients.NSX_DEFENDER_ANALYSIS_URLS
+
     def _login(self) -> None:
         """Implement interface (analysis client relies on 'key' and 'api_token')."""
         if self._session is None:
             self._session = requests.sessions.session()
-        self.post(
-            "authentication", "login", data=tau_clients.purge_none(self._login_params)
-        )
+        self.post("authentication", "login", data=tau_clients.purge_none(self._login_params))
 
     def query_file_hash(self, file_hash: str) -> Dict[str, Any]:
         """
@@ -519,19 +596,17 @@ class AnalysisClient(AbstractClient):
         hash_type = tau_clients.get_hash_type(file_hash)
         if not hash_type:
             raise exceptions.InvalidArgument("Hash type not recognized")
-        params = tau_clients.purge_none(
-            {"hash_value": file_hash, "hash_algorithm": hash_type}
-        )
+        params = tau_clients.purge_none({"hash_value": file_hash, "hash_algorithm": hash_type})
         return self.get("analysis", ["query", "file_hash"], params=params)
 
     def get_analysis_tags(
-        self, uuid: str, allow_datacenter_redirect: bool = None
+        self, uuid: str, allow_datacenter_redirect: Optional[bool] = None
     ) -> Dict[str, Any]:
         """
         Get the analysis tags for an analysis task.
 
         :param str uuid: the unique identifier of the submitted task
-        :param bool allow_datacenter_redirect: whether to allow data center redirection
+        :param bool|None allow_datacenter_redirect: whether to allow data center redirection
         :rtype: dict[str, any]
         :return: a dictionary like:
             {
@@ -592,6 +667,16 @@ class AnalysisClient(AbstractClient):
         params = {"uuid": uuid}
         return self.get("analysis", "get_progress", params=params)
 
+    def get_api_utc_timestamp(self) -> datetime.datetime:
+        """
+        Query the API to get its UTC timestamp.
+
+        :rtype: datetime.datetime
+        :return: Current UTC timestamp according to API
+        """
+        start_info = self.get_completed(after="2039-12-31 23:59:59")
+        return tau_clients.parse_datetime(start_info["before"])
+
     def get_completed(
         self,
         after: Union[datetime.datetime, str],
@@ -603,7 +688,7 @@ class AnalysisClient(AbstractClient):
 
         :param datetime.datetime|str after: request tasks completed after this time
         :param datetime.datetime|str|None before: request tasks completed before this time
-        :param include_score: if True, the response contains score
+        :param include_score: if True, the response contains the score
         :return: a dictionary like:
             {
                 'tasks': ['182a645e7020001000de1474baf8b7b9'],
@@ -612,11 +697,19 @@ class AnalysisClient(AbstractClient):
                 'resume': '2021-07-14 14:09:47',
                 'before': '2021-07-14 14:09:59'
             }
+            or if score is included
+            {
+                'tasks': {'182a645e7020001000de1474baf8b7b9': 70},
+                'after': '2021-07-14 09:10:00',
+                'more_results_available': 0,
+                'resume': '2021-07-14 14:09:47',
+                'before': '2021-07-14 14:09:59'
+            }
         """
         if hasattr(before, "strftime"):
-            before = before.strftime(self.DATETIME_FMT)
+            before = before.strftime(tau_clients.DATETIME_FMT)
         if hasattr(after, "strftime"):
-            after = after.strftime(self.DATETIME_FMT)
+            after = after.strftime(tau_clients.DATETIME_FMT)
         params = tau_clients.purge_none(
             {
                 "before": before,
@@ -626,17 +719,234 @@ class AnalysisClient(AbstractClient):
         )
         return self.get("analysis", "completed", params=params)
 
-    def get_result(self, uuid: str) -> Dict[str, Any]:
+    def get_task_metadata(self, uuid: str) -> Dict[str, str]:
         """
-        Get report results for a given task.
+        Get the metadata of a given task.
 
         :param str uuid: the unique identifier of the submitted task
-        :rtype: dict[str, any]
-        :return: a dictionary containing the analysis report
+        :rtype: dict[str, str]
+        :return: a dictionary like:
+            {
+                'file_md5': '561cffbaba71a6e8cc1cdceda990ead4',
+                'file_mime_type': 'application/x-pe-app-32bit-i386',
+                'file_sha1': '5162f14d75e96edb914d1756349d6e11583db0b0',
+                'file_sha256': 'd55f983c994caa160ec63a59f6b4250fe67fb3e8c43a388aec60a4a6978e9f1e',
+                'filename': 'agent.exe',
+                'task_type': 'file',
+                'task_uuid': 'aba8dccf1641001000c9dd8526ac7c9d',
+            }
         """
         params = {"uuid": uuid}
+        return self.get("analysis", "get_task_metadata", params=params)
+
+    def get_task_sample(self, uuid: str) -> Optional[tau_clients.NamedBytesIO]:
+        """
+        Get the sample analyzed by the provided task if available.
+
+        :param str uuid: the unique identifier of the submitted task
+        :rtype: NamedBytesIO|None
+        :return: the sample
+        """
+        try:
+            artifact_name = "primary_analysis_subject"
+            ret = self.query_artifact(uuid, artifact_name)
+            return self.get_artifact(
+                uuid=uuid,
+                report_uuid=ret["report_uuid"],
+                artifact_name=artifact_name,
+            )
+        except KeyError:
+            return None
+
+    def get_file_sample(self, file_hash: str) -> Optional[tau_clients.NamedBytesIO]:
+        """
+        Get the sample.
+
+        :param str file_hash: the file hash
+        :rtype: NamedBytesIO|None
+        :return: the sample
+        """
+        try:
+            ret = self.query_file_hash(file_hash)
+            for task in ret.get("tasks", []):
+                return self.get_task_sample(task["task_uuid"])
+            else:  # pylint: disable=W0120
+                return None
+        except KeyError:
+            return None
+
+    def query_artifact(
+        self,
+        uuid: str,
+        artifact_name: str,
+        allow_datacenter_redirect: Optional[bool] = None,
+    ) -> Dict[str, str]:
+        """
+        Query if an artifact is available.
+
+        :param str uuid: the unique identifier of the submitted task
+        :param str artifact_name: the artifact name
+        :param bool|None allow_datacenter_redirect: whether to allow data center redirection
+        :rtype: dict[str, str]
+        :return: a dictionary like:
+            {
+                'available': 1,
+                'artifact_name': 'primary_analysis_subject',
+                'task_uuid': 'aba8dccf1641001000c9dd8526ac7c9d',
+                'report_uuid': '19a91b8904dbb7ec29G8gRPpR4dbnAxb1koI7umOuDKVtUkmLX2zyA',
+            }
+        """
+        params = tau_clients.purge_none(
+            {
+                "uuid": uuid,
+                "artifact_name": artifact_name,
+                "allow_datacenter_redirect": allow_datacenter_redirect,
+            }
+        )
+        return self.get("analysis", "query_task_artifact", params=params)
+
+    def get_result(
+        self,
+        uuid: str,
+        report_uuid: Optional[str] = None,
+        include_report: Optional[bool] = True,
+    ) -> Dict[str, Any]:
+        """
+        Get the result for a given task.
+
+        :param str uuid: the unique identifier of the submitted task
+        :param str|None report_uuid: if specified, include this specific report
+        :param bool include_report: whether to include a report
+        :rtype: dict[str, any]
+        :return: the result of an analysis
+        """
+        params = tau_clients.purge_none(
+            {
+                "uuid": uuid,
+                "report_uuid": report_uuid,
+                "full_report_score": 0 if include_report else -1,
+            }
+        )
         return self.get("analysis", "get", params=params)
 
+    def get_result_process_snapshot_names(self, uuid: str) -> List[Dict[str, str]]:
+        """
+        Get all the process snapshot names for a given task.
+
+        :param str uuid: the unique identifier of the submitted task
+        :rtype: list[dict[str, str]]
+        :return: a list of dictionaries like:
+            [
+                {
+                    "task_uuid": "aba8dccf1641001000c9dd8526ac7c9d",
+                    "report_uuid": "4ffa644851ebe118wu1pbwlt3C44cwAYRtfdXY5L3gRhHJ_QANAyuQ",
+                    "artifact_name": "process_snapshots_2",
+                    "artifact_type": "process_snapshot",
+                },
+                ...
+            ]
+        """
+        return self.get_result_artifact_names(
+            uuid=uuid,
+            report_types=[tau_clients.REPORT_TYPE_SANDBOX],
+            metadata_types=[tau_clients.METADATA_TYPE_PROCESS_SNAPSHOT],
+        )
+
+    def get_result_pcap_names(self, uuid: str) -> List[Dict[str, str]]:
+        """
+        Get all the traffic PCAPs names for a given task.
+
+        :param str uuid: the unique identifier of the submitted task
+        :rtype: list[dict[str, str]]
+        :return: a list of dictionaries like:
+            [
+                {
+                    "task_uuid": "aba8dccf1641001000c9dd8526ac7c9d",
+                    "report_uuid": "4ffa644851ebe118wu1pbwlt3C44cwAYRtfdXY5L3gRhHJ_QANAyuQ",
+                    "artifact_name": "traffic.pcap",
+                    "artifact_type": "traffic_capture",
+                },
+                ...
+            ]
+        """
+        return self.get_result_artifact_names(
+            uuid=uuid,
+            report_types=[tau_clients.REPORT_TYPE_SANDBOX],
+            metadata_types=[tau_clients.METADATA_TYPE_PCAP],
+        )
+
+    def get_result_artifact_names(
+        self,
+        uuid: str,
+        report_types: Optional[Iterable[str]] = None,
+        metadata_types: Optional[Iterable[str]] = None,
+    ) -> List[Dict[str, str]]:
+        """
+        Get all the artifact names for a given task.
+
+        :param str uuid: the unique identifier of the submitted task
+        :param iterable[str]|None report_types: optional list of report types to filter on
+        :param iterable[str]|None metadata_types: optional list of metadata types to filter on
+        :rtype: list[dict[str, str]]
+        :return: a list of dictionaries like:
+            [
+                {
+                    "task_uuid": "aba8dccf1641001000c9dd8526ac7c9d",
+                    "report_uuid": "4ffa644851ebe118wu1pbwlt3C44cwAYRtfdXY5L3gRhHJ_QANAyuQ",
+                    "artifact_name": "process_snapshots_2",
+                    "artifact_type": "process_snapshot",
+                },
+                ...
+            ]
+        """
+        ret = []
+        if not report_types:
+            report_types = set([])
+        if not metadata_types:
+            metadata_types = set([])
+        result = self.get_result(uuid, include_report=False)
+        for report in result.get("reports", []):
+            if not report_types or any(x in report["report_versions"] for x in report_types):
+                temp = self.get_result(uuid, report_uuid=report["report_uuid"])
+                for item in temp.get("report", {}).get("analysis_metadata", []):
+                    if not metadata_types or item["metadata_type"] in metadata_types:
+                        ret.append(
+                            {
+                                "task_uuid": uuid,
+                                "report_uuid": report["report_uuid"],
+                                "artifact_name": item["name"],
+                                "artifact_type": item["metadata_type"],
+                            }
+                        )
+        return ret
+
+    def get_artifact(
+        self,
+        uuid: str,
+        report_uuid: str,
+        artifact_name: str,
+        allow_datacenter_redirect: Optional[bool] = None,
+    ) -> tau_clients.NamedBytesIO:
+        """
+        Get an artifact.
+
+        :param str uuid: the unique identifier of the submitted task
+        :param str report_uuid: the selected report
+        :param str artifact_name: the artifact name
+        :param bool|None allow_datacenter_redirect: whether to allow data center redirection
+        :rtype: NamedBytesIO
+        :return: the artifact
+        """
+        params = tau_clients.purge_none(
+            {
+                "uuid": uuid,
+                "report_uuid": report_uuid,
+                "artifact_name": artifact_name,
+                "allow_datacenter_redirect": allow_datacenter_redirect,
+            }
+        )
+        return self.get("analysis", "get_report_artifact", params=params, fmt=None, raw=True)
+
     def submit_file(
         self,
         file_data: bytes,
@@ -654,8 +964,9 @@ class AnalysisClient(AbstractClient):
         :param str|None file_name: if set, represents the name of the file to submit
         :param str|None password: if set, use it to extract the sample
         :param str|None analysis_env: if set, e.g windowsxp
-        :param boolean allow_network_traffic: if set to False, deny network connections
-        :param boolean bypass_cache: whether to re-process a file
+        :param bool allow_network_traffic: if set to False, deny network connections
+        :param bool bypass_cache: whether to re-process a file
+        :param bool include_report: whether to include the report in the result
         :rtype: dict[str, any]
         :return: a dictionary in the following form if the analysis is already available:
             {
@@ -711,8 +1022,8 @@ class AnalysisClient(AbstractClient):
         :param str url: the url to analyze
         :param str|None referer: the referer
         :param str|None user_agent: the user agent
-        :param boolean bypass_cache: bypass_cache
-        :param boolean include_report: whether to include the report in the result
+        :param bool bypass_cache: bypass_cache
+        :param bool include_report: whether to include the report in the result
         :rtype: dict[str, any]
         :return: a dictionary like the following if the analysis is already available:
             {
@@ -747,3 +1058,110 @@ class AnalysisClient(AbstractClient):
         if not include_report:
             ret_data.pop("report", None)
         return ret_data
+
+
+class MultiClientMixin:
+    """
+    This is a mixin that allows any inheriting class to instantiate itself multiple times.
+
+    In other words, we use this to transparently hide that the data returned by a client might
+        depend on the data center being used; using this mixin it is possible to re-define the
+        methods in question and act accordingly by:
+            - calling the method on all instances
+            - merge the results with a bespoke lambda function
+
+    In order to work the class needs to correctly initialize the attribute 'available_clients':
+        - when done correctly the attribute 'available_clients' contains a dictionary of
+            other client objects (including a proxy reference to itself) covering all data centers.
+        - if NOT correctly initialized (via the constructor for example), the instance behaves
+            like a normal client object.
+    """
+
+    @staticmethod
+    def _call_super_method(
+        instance: Any,
+        name: str,
+        *args: Any,
+        **kwargs: Any,
+    ) -> Any:
+        """
+        Call the super of the method 'name' on the instance 'instance'.
+
+        Note: this works because 'super(type, obj)' works the same whether the 'obj' is an actual
+            instance or a proxy to a weakly referenced instance.
+
+        :param MultiClientSubType|ProxyType instance: the instance
+        :param str name: the name of the method to invoke
+        :param args: positional arguments
+        :param kwargs: key-worded arguments
+        :rtype: any
+        :return: what the method 'name' would return
+        """
+        return getattr(super(MultiClientMixin, instance), name)(*args, **kwargs)
+
+    def _call_everywhere(self, method: Callable, *args: Any, **kwargs: Any) -> Dict[str, Any]:
+        """
+        This method calls the parent method of 'method' for each client available.
+
+        :param callable method: the function to call
+        :param args: the list of positional argument
+        :param kwargs: the dictionary of key-worded arguments
+        :rtype: dict[str, any]
+        :return: results indexed by data center
+        """
+        return {
+            data_center: self._call_super_method(client, method.__name__, *args, **kwargs)
+            for data_center, client in self.available_clients.items()
+        }
+
+    def _call_local(self, method: Callable, *args: Any, **kwargs: Any) -> Dict[str, Any]:
+        """
+        This method calls the parent method of 'method'.
+
+        :param callable method: the function to call
+        :param args: the list of positional argument
+        :param kwargs: the dictionary of key-worded arguments
+        :rtype: dict[str, any]
+        :return: results indexed by data center
+        """
+        return {"local": self._call_super_method(self, method.__name__, *args, **kwargs)}
+
+    def __init__(self, *args, **kwargs) -> None:
+        """Constructor."""
+        super(MultiClientMixin, self).__init__(*args, **kwargs)
+        self.available_clients = {}
+
+    @classmethod
+    def from_conf(  # pylint:disable=W0221
+        cls,
+        conf: configparser.ConfigParser,
+        section_name: str,
+        default_data_center: str = tau_clients.NSX_DEFENDER_DC_WESTUS,
+    ) -> "MultiClientSubType":
+        """
+        Get the client from a 'ConfigParser' object.
+
+        :param configparser.ConfigParser conf: the conf object
+        :param str section_name: the section name
+        :param str default_data_center: the default data center
+        :rtype: MultiClientSubType
+        :return: the initialized client
+        :raises configparser.Error: any exception that can be raised by ConfigParser
+        """
+        clients = cls.from_conf_all(conf, section_name)
+        self = clients[default_data_center]
+        clients[default_data_center] = weakref.proxy(self)
+        self.available_clients = clients
+        return self
+
+
+class MultiAnalysisClient(MultiClientMixin, AnalysisClient):
+    """This is a multi-data center analysis client."""
+
+    def query_file_hash(self, file_hash: str) -> Dict[str, Any]:
+        """Method overridden."""
+        if self.available_clients:
+            rest = self._call_everywhere(self.query_file_hash, file_hash)
+        else:
+            rest = self._call_local(self.query_file_hash, file_hash)
+        return tau_clients.merge_dicts(rest.values())
